import os,sys,re,time
from datetime import datetime
from TIR import Transferable_Iterative_Reflection

def check_between_subject(dataset_path):
    pass 


def reflect_from_train_set(dataset_path,result_root,openai_api_key,llm_model_type,reflect_student_num,iter_threshold=3,reflect_sample_ratio=1):
    '''
    The dataset_path should be your TRAIN set
    Your input dataset columns MUST be: ['uid','data_type','xy_type','category','student_id','question_id','concept_id','correctness','question_content','concept_content']
    If your datasets do not have some columns, you still need to put None for these columns.
    
    [uid]: the unique identifier for each sample. Each sample is one question from a specific student. You can understand the uid as the unique index for the table. 
    But you MUST use uuid python library to generate a unique uid per row. You can NOT just use index to be uid.

    [data_type]: this value is either 'train' or 'test'. It determines whether each sample will be used for training or testing. 
    Note that the data_type assignment MUST be between-subject for students (same as pyKT). 
    This means all questions from one student should be all 'train' or all 'test'.
    You can NOT make some questions from one student to be 'train' but some other questions from the same student to be 'test'.

    This means you should first divide students into either train or test group based on a split ratio. After that, you should assign all rows from one student to either
    'train' or 'test'.

    [xy_type]: this value is either 'past' or 'future'. 


    [category]:

    [concept_id]: If a question_id is related to multiple concepts, you should expand it into multiple rows based on the multiple concepts. 
    For example, if question 1 is related to concepts 2,3,4, then you should have three rows: [1,2],[1,3],[1,4] for [question_id,concept_id]. 
    
    However, when you are splitting the dataset into train/test set, you MUST pay attention that all samples related to this question should be put into the same set (either train/test).
    You can NOT put some (such as [1,2]) in train set but some (such as [1,3]) in test set. This will result in label leakage because our evaluation is question-level.

    reflect_sample_ratio: the ratio (0~1) of students used for reflection.  
    reflect_student_num: how many students you want to select for reflection. Note that the students will be selected AFTER filtering students with reflect_sample_ratio.
    iter_threshold: how many iterations for reflection. We recommend it to be 3~5.
    '''
    check_between_subject(dataset_path)
    log_folder = result_root+'/'+datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    os.makedirs(log_folder, exist_ok=True)
    pipeline = Transferable_Iterative_Reflection(llm_model_type=llm_model_type,dataset_path=dataset_path,log_folder=log_folder,print_log=True,openai_api_key=openai_api_key)

    pipeline.dataset_prepare(datasample_flag=True,reflect_sample_ratio=reflect_sample_ratio,random_seed=4)
    pipeline.run_reflect(student_num=reflect_student_num,iter_threshold=iter_threshold)



def reflect_test_set_from_train_set(dataset_path,result_root,openai_api_key,llm_model_type,example_student_num_per_category,test_student_num_per_category,iter_threshold=3):
    check_between_subject(dataset_path)
    log_folder = result_root+'/'+datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    os.makedirs(log_folder, exist_ok=True)
    pipeline = Transferable_Iterative_Reflection(llm_model_type=llm_model_type,dataset_path=dataset_path,log_folder=log_folder,print_log=True,openai_api_key=openai_api_key)

    pipeline.run_predict_reuse_reflect(example_student_num_per_course=example_student_num_per_category,test_student_num_per_course=test_student_num_per_category,random_seed=4,iter_threshold=iter_threshold)




# calculate_metrics('/home/songlin/study_results/GKT/reflection_generation/direct_reflect_reuse_history_all_8past/predict_post_detail.csv')


def run_exp_bertLLMKT(dataset_train_path,dataset_test_path,log_folder):
    experiment_pipeline = Experiment_Pipeline(512,log_folder,load_model_type='none',lr=5e-7)
    experiment_pipeline.dataset_prepare(dataset_train_path,dataset_test_path,batch_size=64)
    experiment_pipeline.model_train(epochs=50,vis=True)
    experiment_pipeline.model_eval(eval_mode='test')

    experiment_pipeline = Experiment_Pipeline(512,log_folder,load_model_type='best',lr=5e-7)
    experiment_pipeline.dataset_prepare(dataset_train_path,dataset_test_path,batch_size=64)
    experiment_pipeline.model_eval(eval_mode='test')



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="minihuman",
    version="0.1.2",
    author="human.ai.integration",
    author_email="human.ai.integration@gmail.com",
    description="MiniHuman: A Python Library to Simulate Virtual Human Behaviors and Responses to Environmental Stimuli",
    long_description="MiniHuman: A Python Library to Simulate Virtual Human Behaviors and Responses to Environmental Stimuli",
    long_description_content_type="text/markdown",
    index="https://github.com/songlinxu/MiniHuman-Toolkit",
    project_urls={
        "Bug Tracker": "https://github.com/songlinxu/MiniHuman-Toolkit",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    python_requires=">=3.5",
    include_package_data=True,
    install_requires=['numpy>=1.17.2','pandas>=1.1.5','scikit-learn'],
)


import logging
# from transformers import logging as transformers_logging
# transformers_logging.set_verbosity_error()

import torch
import torch.nn as nn
import torch.optim as optim
from transformers import BertTokenizer, BertModel, EncoderDecoderModel, ReformerModelWithLMHead, ReformerTokenizer, BertForSequenceClassification
from torch.utils.data import DataLoader, Dataset, random_split
import pandas as pd
import numpy as np 
from matplotlib import pyplot as plt 
import os,sys,random,time,shutil,warnings

from sklearn.metrics import accuracy_score, f1_score

def remove_folder(folder_path):
    try:
        shutil.rmtree(folder_path)  # Use shutil.rmtree to remove the directory and its contents
        print(f"Folder '{folder_path}' and all its contents have been removed successfully.")
    except FileNotFoundError:
        print(f"The folder '{folder_path}' does not exist.")
    except Exception as e:
        print(f"An error occurred while removing the folder: {e}")



class CustomDataset(Dataset):
    def __init__(self, dataframe, tokenizer, max_length):
        self.dataframe = dataframe
        self.tokenizer = tokenizer
        self.max_length = max_length

        if tokenizer.pad_token_id is None: tokenizer.add_special_tokens({'pad_token': '[PAD]'})

    def __len__(self):
        return len(self.dataframe)

    def __getitem__(self, idx):
        input_prompt = self.dataframe.iloc[idx]['input_prompt']
        # input_prompt_1 = self.dataframe.iloc[idx]['input_past_prompt']
        # input_prompt_2 = self.dataframe.iloc[idx]['input_future_prompt']
        # input_prompt_3 = self.dataframe.iloc[idx]['initial_response']
        # input_prompt_4 = self.dataframe.iloc[idx]['best_reflection']
        label = self.dataframe.iloc[idx]['future_label']
        
        # inputs = self.tokenizer(input_prompt_2, input_prompt_3, input_prompt_4, truncation='longest_first', padding='max_length', max_length=self.max_length, return_tensors='pt')
        # inputs = self.tokenizer(input_prompt_3, truncation='longest_first', padding='max_length', max_length=self.max_length, return_tensors='pt')
        inputs = self.tokenizer(input_prompt_1, input_prompt_2, truncation='longest_first', padding='max_length', max_length=self.max_length, return_tensors='pt')

        item = {key: val.squeeze() for key, val in inputs.items()}
        item['labels'] = torch.tensor(label, dtype=torch.long)
        # print('input_token_len: ',len(self.tokenizer.tokenize(input_prompt_1))+len(self.tokenizer.tokenize(input_prompt_2)))

        return item


class lm_classifier_pipeline():
    def __init__(self, max_length, log_folder, load_model_type, random_seed = 4, lr=1e-4):
        self.max_length = max_length
        self.set_seed(random_seed)

        self.log_folder = log_folder
        self.load_model_type = load_model_type

        self.model_init(load_model_type,lr)

    def set_seed(self,seed_num):
        np.random.seed(seed_num)
        random.seed(seed_num)
        torch.manual_seed(seed_num)
        torch.cuda.manual_seed(seed_num)
        torch.cuda.manual_seed_all(seed_num)  
        torch.backends.cudnn.deterministic = True 
        torch.backends.cudnn.benchmark = False

    def dataset_prepare(self,dataset_train_path,dataset_test_path,batch_size = 16):
        dataframe_train = pd.read_csv(dataset_train_path,sep='\t')
        dataframe_test = pd.read_csv(dataset_test_path,sep='\t')
        
        dataset = CustomDataset(dataframe_train, self.tokenizer, self.max_length)
        test_dataset = CustomDataset(dataframe_test, self.tokenizer, self.max_length)
        
        train_size = int(0.8 * len(dataset))
        val_size = len(dataset) - train_size 
        train_dataset, val_dataset = random_split(dataset, [train_size, val_size])
        
        self.train_dataloader = DataLoader(train_dataset, shuffle=True, batch_size=batch_size)
        self.val_dataloader = DataLoader(val_dataset, batch_size=batch_size)
        self.test_dataloader = DataLoader(test_dataset, batch_size=batch_size)


    def model_save(self,checkpoint,checkpoint_path,tokenizer,tokenizer_path,e2emodel,e2emodel_path):
        if os.path.exists(checkpoint_path): os.remove(checkpoint_path)
        torch.save(checkpoint, checkpoint_path)
        if os.path.exists(tokenizer_path): remove_folder(tokenizer_path)
        tokenizer.save_pretrained(tokenizer_path)
        if os.path.exists(e2emodel_path): remove_folder(e2emodel_path)
        e2emodel.save_pretrained(e2emodel_path)


    def model_init(self,load_model_type,lr=1e-4):
        assert load_model_type in ['best','last','none']

        self.checkpoint_last_path = self.log_folder + '/model_last.pt'
        self.checkpoint_best_path = self.log_folder + '/model_best.pt'
        self.tokenizer_last_path = self.log_folder + '/tokenizer_last'
        self.tokenizer_best_path = self.log_folder + '/tokenizer_best'
        self.e2emodel_last_path = self.log_folder + '/e2emodel_last'
        self.e2emodel_best_path = self.log_folder + '/e2emodel_best'

        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        print(self.device)

        if load_model_type in ['best','last']:
            self.tokenizer = BertTokenizer.from_pretrained(self.tokenizer_last_path) if load_model_type == 'last' else BertTokenizer.from_pretrained(self.tokenizer_best_path)
            self.model = BertForSequenceClassification.from_pretrained(self.e2emodel_last_path).to(self.device) if load_model_type == 'last' else BertForSequenceClassification.from_pretrained(self.e2emodel_best_path).to(self.device)
            checkpoint = torch.load(self.checkpoint_last_path) if load_model_type == 'last' else torch.load(self.checkpoint_best_path)
            self.epoch_exist = checkpoint['epoch']
            # self.model.load_state_dict(checkpoint['model_state_dict'])
            self.optimizer = optim.AdamW(self.model.parameters(), lr=lr)
            self.optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
            self.train_losses = checkpoint['train_loss']
            self.val_losses = checkpoint['val_loss']
        else:
            self.tokenizer = BertTokenizer.from_pretrained('bert-base-uncased')
            self.model = BertForSequenceClassification.from_pretrained('bert-base-uncased', num_labels=2).to(self.device)
            self.optimizer = optim.AdamW(self.model.parameters(), lr=lr)
            self.epoch_exist = 0
            self.train_losses = []
            self.val_losses = []

        # self.model.config.decoder_start_token_id = self.tokenizer.cls_token_id
        # self.model.config.pad_token_id = self.tokenizer.pad_token_id

        self.criterion = nn.CrossEntropyLoss()
        # self.criterion = nn.CrossEntropyLoss(ignore_index=self.tokenizer.pad_token_id)

    def model_train(self, epochs=3, vis=True):
        loss_file = self.log_folder+'/loss.csv'
        with open(loss_file, "a+") as file1:
            file1.write('epoch,train_loss,val_loss,accuracy,f1\n')

        for epoch in range(epochs):
            if epoch + 1 <= self.epoch_exist: continue
            self.model.train()
            total_loss = 0
            time_train_start = time.time()
            for batch in self.train_dataloader:
                self.optimizer.zero_grad()
                input_ids = batch['input_ids'].to(self.device)
                attention_mask = batch['attention_mask'].to(self.device)
                labels = batch['labels'].to(self.device)

                # outputs = self.model(
                #     input_ids=input_ids,
                #     attention_mask=attention_mask,
                #     labels=labels
                # )

                # print('outputs: ',outputs)

                # loss = outputs.loss

                outputs = self.model(input_ids, attention_mask)
                loss = self.criterion(outputs.logits, labels)
                # loss = self.criterion(outputs.view(-1, outputs.size(-1)), labels.view(-1))
                loss.backward()
                self.optimizer.step()
                
                total_loss += loss.item()
            time_train_end = time.time()

            avg_loss = total_loss / len(self.train_dataloader)
            self.train_losses.append(avg_loss)
            train_time = time_train_end - time_train_start
            print(f'Epoch {epoch + 1}, Training loss: {avg_loss}, Training Time: {train_time}')
            
            val_loss, accuracy, f1 = self.model_eval(eval_mode='val')

            with open(loss_file, "a+") as file1:
                file1.write(str(epoch + 1)+','+str(avg_loss)+','+str(val_loss)+','+str(accuracy)+','+str(f1)+'\n')

            checkpoint = {
                'epoch': epoch + 1,
                # 'model_state_dict': self.model.state_dict(),
                'optimizer_state_dict': self.optimizer.state_dict(),
                'train_loss': self.train_losses,
                'val_loss': self.val_losses
            }
            
            self.model_save(checkpoint,self.checkpoint_last_path,self.tokenizer,self.tokenizer_last_path,self.model,self.e2emodel_last_path)

            if (len(self.val_losses) == 0) or (len(self.val_losses) != 0 and val_loss == min(self.val_losses)):
                self.model_save(checkpoint,self.checkpoint_best_path,self.tokenizer,self.tokenizer_best_path,self.model,self.e2emodel_best_path)

        if vis == True:
            fig, ax = plt.subplots()
            plt.plot(self.train_losses, label='Train Loss')
            plt.plot(self.val_losses, label='Validation Loss')
            plt.xlabel("Epoch")
            plt.ylabel("Loss")
            plt.title("Training and Validation Loss")
            plt.legend()
            plt.savefig(self.log_folder+'/loss.png')

    def model_eval(self,eval_mode):
        self.model.eval()
        total_loss = 0
        all_preds = []
        all_labels = []
        
        assert eval_mode in ['test','val']
        dataloader_part = self.test_dataloader if eval_mode == 'test' else self.val_dataloader
        time_eval_start = time.time()

        with torch.no_grad():
            for batch in dataloader_part:
                input_ids = batch['input_ids'].to(self.model.device)
                attention_mask = batch['attention_mask'].to(self.model.device)
                labels = batch['labels'].to(self.model.device)

                outputs = self.model(input_ids, attention_mask)
                # outputs = self.model(
                #     input_ids=input_ids,
                #     attention_mask=attention_mask,
                #     labels=labels
                # )
                # loss = outputs.loss
                loss = self.criterion(outputs.logits, labels)
                # loss = self.criterion(outputs.view(-1, outputs.size(-1)), labels.view(-1))
                total_loss += loss.item()

                # Calculate accuracy
                preds = torch.argmax(outputs.logits, dim=1)
                all_preds.extend(preds.cpu().numpy())
                all_labels.extend(labels.cpu().numpy())

        time_eval_end = time.time()
        time_eval = time_eval_end - time_eval_start
        avg_loss = total_loss / len(dataloader_part)
        self.val_losses.append(avg_loss)

        accuracy = accuracy_score(all_labels, all_preds)
        f1 = f1_score(all_labels, all_preds, average='weighted')

        print(f'{eval_mode} loss: {avg_loss}, accuracy: {accuracy}, F1 score: {f1}, time: {time_eval}')
        return avg_loss, accuracy, f1



